  '==========================================================================================================
  '
  '       FILE UTILS - Version 1.0.3  (23 Nov 2021)
  '
  '                    Written by: Phil Erb
  '
  '==========================================================================================================

  '           1         2         3         4 
  '  1234567890123456789012345678901234567890
  ?                                               '0  
  ? "The FileUtils program is not intended to"    '1
  ? "be executed directly. It contains a"         '2 
  ? "library of functions that are designed"      '3
  ? "to be standalone and executed from the"      '4
  ? "Command Prompt. The FileUtils program"       '5
  ? "only needs to be LOADED not RUN."            '6
  ?                                               '7
  ? "These functions provide utilities for"       '8
  ? "managing and working with files on a"        '9 
  ? "SD Card."                                    '10
  ?                                               '11
  ? "To view more information about FileUtils"    '12
  ? "execute the FileUtilsHelp function by"       '13
  ? "entering "+ Chr$(34) + "FileUtilsHelp()";
  ?  Chr$(34) + " (without the"                   '14
  ? "double quotes) at the Command Prompt."       '15  
  ? "E.G.   >FileUtilsHelp()"                     '16

  If WaitContinue%(0) Then                        'need to wait for ok to continue
  EndIf  

  ?
  ? "Program terminated - functions are still"
  ? "accessible from the Command Prompt."

  END

  
  '==========================================================================================================
  '
  '       FILE UTILS HELP Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileUtilsHelp(spWhich$)
    'This routine displays HELP information for the functions supported by the FileUtils library.
    'Help information format assumes 320x240 screen using Font #1.
    If spWhich$ = "" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? " File Utils Library HELP Information - 1"    '1
      ? " ======================================="    '2
      ? "Functions supporting record based files"     '3
      ? "using CR/LF line terminators."               '4
      ?                                               '5
      ? "  FileCopy - Copies one file as another"     '6
      ? "    file name."                              '7
      ?                                               '8
      ? "  FileExtract - Copies a specified part"     '9
      ? "    of one file as another file name."       '10
      ?                                               '11
      ? "  FileList - Lists the contents of the"      '12
      ? "    specified file (printable files)."       '13
      ?                                               '14
      ? "  FileMerge - Merges one file into"          '15
      ? "    another at a specified position."        '16

      If WaitContinue%(0) Then                        'need to wait for ok to continue
      EndIf  
      
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? " File Utils Library HELP Information - 2"    '1
      ? " ======================================="    '2
      ? "Functions supporting stream (binary)"        '3
      ? "files - i.e. no CR/LF line terminators."     '4
      ?                                               '5
      ? "  FileBinCopy - Copies one file as"          '6
      ? "    another file name."                      '7
      ?                                               '8
      ? "  FileBinList - Lists the contents of"       '9
      ? "    the specified file in Hex and alpha."    '10
      ?                                               '11
      ? "  FileBinXfer - Allows copying a file"       '12
      ? "    from one SD Card to a different"         '13
      ? "    card by swapping the SD Card when"       '14
      ? "    requested to do so."                     '15
      ?                                               '16
      ? "Use FileUtilsHelp(<func$>) for more info"    '17 
      
    ElseIf UCase$(spWhich$) = "FILECOPY" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileCopy HELP Information"             '1
      ? "      ========================="             '2
      ? "This function copies one file as another"    '3
      ? "file name, leaving the original file"        '4
      ? "un-touched."                                 '5
      ?                                               '6
      ? "Syntax: FileCopy(<srcfile$>,<destfile$>)"    '7
      ?                                               '8 
      ? "Copies the contents of the specified"        '9 
      ? "<srcfile> to the <destfile>, replacing"      '10
      ? "the <destfile> if it already exists."        '11
      ? "Both <srcfile$> and <destfile$> can"         '12
      ? "include a path specification, and if"        '13 
      ? "the dest path doesn't exist it will be"      '14
      ? "created."                                    '15
      ? "E.G. -"                                      '16
      ? "FileCopy("+ Chr$(34) + "A:/File.bas" + Chr$(34) + ",";
      ? Chr$(34) + "A:/MM/Test.bas" + Chr$(34) +")"   '17
  
    ElseIf UCase$(spWhich$) = "FILEEXTRACT" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "     FileExtract HELP Information"           '1
      ? "     ============================"           '2
      ? "This function copies part of one file as"    '3
      ? "another file name."                          '4
      ?                                               '5
      ? "Syntax: FileExtract(<srcfile$>,<strt%>,"     '6
      ? "                    <end%>,<destfile$>)"     '7
      ?                                               '8  
      ? "Same as FileCopy, except that <strt%>"       '9 
      ? "specifies the record number in <srcfile>"    '10
      ? "to start copying from and <end%>"            '11
      ? "specifies the last record number to be"      '12
      ? "copied. <end%> can be 9999 which will"       '13  
      ? "copy to the end of the source file."         '14
      ? "E.G. -"                                      '15
      ? "FileExtract(" + Chr$(34) + "A:/File.bas" + Chr$(34) + ",25,68,"  '16
      ? "            " + Chr$(34) + "A:/MM/Test.bas" + Chr$(34) +")"      '17
  
    ElseIf UCase$(spWhich$) = "FILELIST" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileList HELP Information"             '1
      ? "      ========================="             '2
      ? "This function lists the contents of the"     '3
      ? "specified file (.bas,.txt, etc files)."      '4
      ?                                               '5
      ? "Syntax: FileList(<filename$>)"               '6
      ?                                               '7  
      ? "<filename$> can include a path"              '8 
      ? "specification. The listing will be"          '9
      ? "formatted for 40 chars/line and 18 lines"    '10
      ? "per console screen and will display"         '11
      ? "record numbers for reference. "              '12 
      ? "The listing will pause with the"             '13
      ? "'Press any key to continue...' message"      '14  
      ? "when the console screen is filled."          '15
      ?                                               '16
      ? "E.G. - FileList(" + Chr$(34) + "A:/Test/MyFile.txt" + Chr$(34) + ")" '17
  
    ElseIf UCase$(spWhich$) = "FILEMERGE" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileMerge HELP Information"            '1
      ? "      =========================="            '2
      ? "This function inserts (merges) one file"     '3
      ? "into another file at a specified posn."      '4
      ?                                               '5
      ? "Syntax: FileMerge(<insertfile$>,"            '6
      ? "                  <mergefile$>,<posn%>)"     '7
      ?                                               '8  
      ? "Inserts the content of <insertfile> into"    '9 
      ? "<mergefile> starting AFTER the record"       '10
      ? "number specified by <posn%>. The entire"     '11
      ? "<insertfile> content is inserted,"           '12
      ? "followed by the remaining records of"        '13  
      ? "the original <mergefile> (if any)."          '14
      ? "E.G. -"                                      '15
      ? "FileMerge(" + Chr$(34) + "A:/File.bas" + Chr$(34) + ","          '16
      ? "          " + Chr$(34) + "MM/TestMerge.bas" + Chr$ (34) +",25)"  '17
      
    ElseIf UCase$(spWhich$) = "FILEBINCOPY" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileBinCopy HELP Information"          '1
      ? "      ============================"          '2
      ? "This function copies one file as another"    '3
      ? "file name, leaving the original file"        '4
      ? "un-touched."                                 '5
      ?                                               '6
      ? "Syntax: FileBinCopy(<srcfile$>,"             '7    
      ? "                    <destfile$>)"            '8
      ?                                               '9  
      ? "Same as FileCopy except that it handles"     '10 
      ? "non-record based (binary) files."            '11
      ? "See FileCopy for more details."              '12
      ? "Can also be used for record based files."    '13  
      ?                                               '14
      ? "E.G. -"                                      '15
      ? "FileBinCopy("+ Chr$(34) + "A:/File.bas" + Chr$(34) + ","   '16
      ? "            "+ Chr$(34) + "A:/MM/Test.bas" + Chr$(34) +")" '17
      
    ElseIf UCase$(spWhich$) = "FILEBINLIST" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileBinList HELP Information"          '1
      ? "      ============================"          '2
      ? "This function lists the contents of the"     '3
      ? "specified file in both Hexadecimal and "     '4
      ? "alphanumeric (for printable characters)."    '5
      ?                                               '6
      ? "Syntax: FileBinList(<filename$>,<strt%>)"    '7
      ?                                               '8  
      ? "Same as FileList except the output is in"    '9
      ? "hex and alpha format showing 8 bytes per"    '10
      ? "line. <strt%>, if specified, is the byte"    '11
      ? "offset the listing will start from. The"     '12 
      ? "default is 0 which is the first byte."       '13
      ? "The listing will pause with the"             '14
      ? "'Press any key to continue...' message"      '15  
      ? " when the console screen is filled."         '16
      ? "E.G. - FileBinList(" + Chr$(34) + "A:/MyFile.txt" + Chr$(34) + ",861)" '17
      
    ElseIf UCase$(spWhich$) = "FILEBINXFER" Then
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "      FileBinXfer HELP Information"          '1
      ? "      ============================"          '2
      ? "This function copies a file from one"        '3
      ? "SD Card to a different SD Card. It can"      '4
      ? "be used for both binary and record based"    '5
      ? "files."                                      '6
      ?                                               '7
      ? "Syntax: FileBinXfer(<srcfile$>,"             '8    
      ? "                    <destfile$>)"            '9
      ?                                               '10  
      ? "Requires the user to swap SD Cards when"     '11
      ? "requested to do so. The destination file"    '12
      ? "MUST be specified with the full path"        '13  
      ? "starting from 'A:/' (root directory)."       '14
      ? "E.G. -"                                      '15
      ? "FileBinCopy("+ Chr$(34) + "File.bas" + Chr$(34) + ","      '16
      ? "            "+ Chr$(34) + "A:/MM/Test.bas" + Chr$(34) +")" '17
     
    Else
    
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ?                                               '0
      ? "  File Utils library HELP Information"       '1 
      ? "  ==================================="       '2
      ?                                               '3
      ? "Help requested for an unknown function:"     '4
      ?                                               '5
      ? "   " + spWhich$                              '6
      ?                                               '7
      ? "Check for an incorrect or mis-spelt"         '8 
      ? "function name."                              '9
      ?                                               '10
      ?                                               '11
      ?                                               '12
      ?                                               '13  
      ?                                               '14
      ?                                               '15
      ?                                               '16
      ?                                               '17  
    
    EndIf
    
  End Sub
  

  
  '==========================================================================================================
  '
  '       COPY FILE Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileCopy(spSFN$,spDFN$)
    'This routine copies one file as another file name, leaving the original file un-touched.
    'Both Source File Name and Destination File Name can include Path specifications.
    'If the Path specified for the Destination File does not exist, it will be automatically created.
    
    'This routine uses the FileExtract procedure to do the actual work.
    FileExtract(spSFN$,1,9999,spDFN$,1)               'copy the entire source file   
    
  End Sub   'FileCopy()

  
  '==========================================================================================================
  '
  '       EXTRACT FILE Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileExtract(spSFN$,ipStrt%,ipEnd%,spDFN$,ipAction%)
    'This routine copies a portion of one file as another file name, leaving the original file un-touched.
    'Both Source File Name and Destination File Name can include Path specifications.
    'If the Path specified for the Destination File does not exist, it will be automatically created.
    'ipStrt specifies the record number in the source file to start copying from.
    'ipEnd specifies the last record number to be included in the copy.
    'To begin copy from the start of the source file use ipStrt = 1.
    'To copy to the end of the source file, use ipEnd as either the actual last record number or 9999.
    'ipAction% argument is only used internally for the call from FileCopy to this routine.    
    
    Local b%,n%,m%,a1$,a2$,d$,cp$,fullSFN$,fullDFN$,saveCWD$
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    
    If ipAction% Then                                 'setup action descriptios
      a1$="Cannot copy"
      a2$="copied"
    Else
      a1$="Cannot extract"
      a2$="extracted"
    EndIf
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spSFN$,fullSFN$,cp$)                     'format file name & build full file name
    If Not OpenSFN%(spSFN$,fullSFN$) Then             'source file failed to open -
      Exit Sub                                        'exit now
    EndIf
    
    If LOF(#1) = 0 Then
      ? a1$;" file - Source File is empty"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    FormatFN(spDFN$,fullDFN$,cp$)                     'format file name & build full file name)

    If Len(spDFN$) < 1 Then
      ? a1$;" file - Destination File Name is blank"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    If UCase$(fullSFN$) = UCase$(fullDFN$) Then
      ? a1$;" file - Source and Destination files are the same"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    If ipEnd% < ipStrt% Then
      ? "Invalid record range - Start record must be less than End record"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    'validate and/or build the required directory structure for the dest file
    If Not Validate_Build_DestPath%(spDFN$) Then
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub                                        'build failed - exit now
    EndIf
    
    CHDIR saveCWD$                                    'restore current path
    
    'open destination file
    ON ERROR SKIP 1
    OPEN spDFN$ FOR OUTPUT AS #2
    If MM.ERRNO <> 0 Then                             'file open failed - report error
      ? "Open destination file '";fullDFN$;"' failed - ";
      ? MM.ERRMSG$
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    'copy specified records from source file to destination file
    n%=0
    m%=0
    
    Do
      If Not ReadNextRecord%(fullSFN$,d$) Then Exit Sub
      Inc n%                                          'increment record count read
      If n% >= ipStrt% And n% <= ipEnd% Then          'this is a record we want      
        Print #2,d$                                   '- copy it
        Inc m%                                        'increment record count written
      EndIf
    Loop Until EOF(#1) Or n% >= ipEnd%
    
    'all done - close files and report results
    b%=LOF(#2)                                        'save extract file len in bytes
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    
    If m% = 0 Then
      Kill spDFN$
      ? "Dest file not created - no records "; 
      ? "found within the specified range: ";
      ? ipStrt%;" to ";ipEnd%;
    Else
      ? m%;" records ";a2$;" from ";fullSFN$;" to ";fullDFN$;
      ? " (";a2$;" bytes = "b%;")"
    EndIf
    
  End Sub   'FileExtract()

  
  '==========================================================================================================
  '
  '       LIST FILE Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileList(spSFN$)
    'This routine lists the contents of the specified file.
    'The File Name can include Path specifications.
    
    Local a%,i%,ln%,ln_max%,rec%,ch_ptr%,ch_max%,cp$,disp_buff$,fullSFN$,rec_buff$,saveCWD$
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spSFN$,fullSFN$,cp$)                     'format file name & build full file name
    If Not OpenSFN%(spSFN$,fullSFN$) Then             'source file failed to open -
      Exit Sub                                        'exit now
    EndIf

    'since there is currently no way to correctly determine the OPTION DISPLAY settings
    'applying to the console, I am just using 18 lines * 40 chars per line for formatting
    'console output.
    '18*40 is the correct setting for a ILI9314 LCD using OPTION LCDPANEL CONSOLE 1. 
    'Note that ln_max% is 18 - 2, reserving 1 line for the "Press any key to continue..." msg 
    'plus 1 extra to cater for the way the console line scrolling works.
    
    ch_max%=40
    ln_max%=16
    rec%=1
    ln% = 0
    ?                                                 'display blank line before listing starts
    Do
      disp_buff$=""                                   'clear display buffer
      If Not ReadNextRecord%(fullSFN$,rec_buff$) Then Exit Sub
      For i%=1 to Len(rec_buff$)                      'replace non-printables with space
        a%=Asc(Mid$(rec_buff$,i%,1))
        If a% < 32 Or a% > 126 Then a%=32             'substitue non-printables
        disp_buff$ = disp_buff$ + Chr$(a%)            'build display buffer
      Next i%
      ? Str$(rec%,4) + " ";                           'display record #
      ? Left$(disp_buff$,ch_max% - 5)                 'display record content (maybe only first part)
      inc rec%                                        'increment record #                                       
      inc ln%                                         'increment line # on this screen
      If Len(disp_buff$) > ch_max% - 5 Then           'record longer than ch_max% - 5, need to continue
        ch_ptr% = ch_max% - 4                         'set pointer to next char to display from rec_buff$
        Do
          If ln% >= ln_max% Then                      'screen full
            If WaitContinue%(1) Then                  '- need to wait for ok to continue or Esc to quit
              Exit Do
            EndIf
            ?                                         'add blank line at top
            ln% = 0                                   'reset line count
          EndIf
          ? Mid$(disp_buff$,ch_ptr%,ch_max%)          'display next part of record content
          inc ln%                                     'increment line # on this screen
          inc ch_ptr%,ch_max%                         'increment ch_ptr% by number of chars just displayed      
        Loop Until ch_ptr% >= Len(rec_buff$)          'loop until all the record content has been displayed
      EndIf
      If ln% >= ln_max% Then                          'screen full
        If WaitContinue%(1) Then                      '- need to wait for ok to continue or Esc to quit
          Exit Do
        EndIf
        ?                                             'add blank line at top
        ln% = 0                                       'reset line count
      EndIf
    Loop Until EOF(#1)
    
    ?                                                 'display blank line
    ? (rec%-1);" records listed from '";fullSFN$;"'."
    
    'all done - close file
    ON ERROR SKIP 1
    CLOSE #1
    
  End Sub   'FileList()

  
  '==========================================================================================================
  '
  '       MERGE FILE Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileMerge(spInsFN$,spMrgFN$,ipPosn%)
    'This routine inserts (merges) one file into another file, starting after a specified record
    'number in the second file (the merge file). The merge file is modified by this function,
    'becoming an amalgamation of the original merge file and the insert file. The insert file 
    'remains un-touched.
    'ipPosn specifies the record number in the original merge file AFTER which the contents of
    'the insert file will be inserted, then the remaining records of the original merge file
    'will follow.
    'If ipPosn is > the number of records in the original merge file, the insert file will be 
    'appended to the end of the merge file (ipPosn can be specified as 9999 to achieve this).
    
    Local f%,i%,n%,m%,c$,d$,cp$,dp$,workFN$,fullInsFN$,fullMrgFN$,saveCWD$
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    ON ERROR SKIP 1
    CLOSE #3
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spInsFN$,fullInsFN$,cp$)                 'format file name & build full file name
    If Not OpenSFN%(spInsFN$,fullInsFN$) Then         'insert file failed to open -
      Exit Sub                                        'exit now
    EndIf
    
    If LOF(#1) = 0 Then
      ? "Cannot merge files - Insert File is empty"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    FormatFN(spMrgFN$,fullMrgFN$,cp$)
    
    If Len(spMrgFN$) < 1 Then
      ? "Cannot merge files - Merge File Name is blank"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    If UCase$(fullInsFN$) = UCase$(fullMrgFN$) Then
      ? "Cannot merge files - Insert file and Merge file are the same"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf

    'determine the path for the Merge file (if path is specified)
    n%=Len(spMrgFN$)
    
    FOR i%=n% TO 1 Step -1
      c$=Mid$(spMrgFN$,i%,1) 
      'check for a fwd slash
      If c$ = "/" Then
        'found
        dp$=Left$(spMrgFN$,i%)                        'get directory name specified
        EXIT FOR
      EndIf
    Next i%

    'setup temporary work file name, in the same directory as the merge file
    workFN$ = dp$ + "tempmerge_" + Str$(EPOCH(NOW)) + ".tmp"
    
    'open remaining 2 files (file #1 - insert file - already opened)
    ON ERROR SKIP 1
    OPEN spMrgFN$ FOR INPUT AS #2
    If MM.ERRNO <> 0 Then                             'file open failed - report error
      ? "Open Merge file '";fullMrgFN$;"' failed - ";MM.ERRMSG$
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    If LOF(#2) = 0 Then
      ? "Cannot merge files - Merge File is empty"
      ON ERROR SKIP 1
      CLOSE #1
      ON ERROR SKIP 1
      CLOSE #2
      Exit Sub
    EndIf
    
    ON ERROR SKIP 1
    OPEN workFN$ FOR OUTPUT AS #3
    If MM.ERRNO <> 0 Then                             'file open failed - report error
      If Instr(workFN$,":") = 0 Then                  'not absolute path for merge file
        workFN$ = cp$ + workFN$                       'build full path 
      EndIf
      ? "Open work file '";workFN$;"' failed - ";MM.ERRMSG$
      ON ERROR SKIP 1
      CLOSE #1
      ON ERROR SKIP 1
      CLOSE #2
      Exit Sub
    EndIf
    
    'copy records from Merge file to work file, until the specified <ipsStrt>
    'records have been copied, then copy all records from the Insert file
    'and finally copy remaining records (if any) from the original Merge file.
    'The work file will then contain the merged contents of the two input files.
    f%=0
    n%=0
    m%=0
    
    Do
      'cannot use ReadNextRecord(..) here as it only reads fnbr# 1, not #2
      ON ERROR SKIP 1
      LINE INPUT #2,d$                                'read next record of original Merge file
      If MM.ERRNO <> 0 Then                           'read failed    
        ? "Read merge file failed - ";MM.ERRMSG$
        ? "File '";fullMrgFN$;"' may be a stream file."
        ON ERROR SKIP 1
        CLOSE #1
        ON ERROR SKIP 1
        CLOSE #2
        ON ERROR SKIP 1
        CLOSE #3
        Exit Sub
      EndIf
      Inc m%                                          'increment record count read (Merge file)
      Print #3,d$                                     'write it to the work file
      If NOT f% Then                                  'check if ready to insert yet
        If m% = ipPosn% Then                          'start the insert now
          Do
            If Not ReadNextRecord%(fullInsFN$,d$) Then Exit Sub
            Inc n%                                    'increment record count read (Insert file)
            Print #3,d$                               'write it to the work file
          Loop Until EOF(#1)                          'loop here until all insert records are processed
          f%=1                                        'set Insert Done flag
        EndIf
      EndIf
    Loop Until EOF(#2)                                'loop until all original Merge file records copied
    
    If f% = 0 Then                        
      'insert was not done, presumably because ipPosn is > # of records in the Merge file,
      'so insert onto the end of the original Merge file
      Do
        If Not ReadNextRecord%(fullInsFN$,d$) Then Exit Sub
        Inc n%                                        'increment record count read (Insert file)
        Print #3,d$                                   'write it to the work file
      Loop Until EOF(#1)                              'loop here until all insert records are processed
    EndIf
    
    'all done - close files and report results
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    ON ERROR SKIP 1
    CLOSE #3
    
    Kill spMrgFN$                                     'delete the original Merge file  
    Pause 500
    Name workFN$ AS spMrgFN$                          'rename the work file as the new Merge file
  
    ? n%;" records merged from ";fullInsFN$;" into ";fullMrgFN$;
    ? " (new record count = ";(m%+n%);")."
    
  End Sub   'FileMerge()
  
  
  '==========================================================================================================
  '
  '       BINARY FILE COPY Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileBinCopy(spSFN$,spDFN$)
    'This routine copies one file as another file name, leaving the original file un-touched.
    'Both Source File Name and Destination File Name can include Path specifications.
    'If the Path specified for the Destination File does not exist, it will be automatically created.
        
    Local e%,f_len%,n%,cp$,d_buff$,fullSFN$,fullDFN$,saveCWD$
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spSFN$,fullSFN$,cp$)                     'format file name & build full file name
    If Not OpenSFN%(spSFN$,fullSFN$) Then             'source file failed to open -
      Exit Sub                                        'exit now
    EndIf
    
    If LOF(#1) = 0 Then
      ? "Cannot copy file - Source File is empty"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    FormatFN(spDFN$,fullDFN$,cp$)
    
    If Len(spDFN$) < 1 Then
      ? "Cannot copy file - Destination File Name is blank"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    If UCase$(fullSFN$) = UCase$(fullDFN$) Then
      ? "Cannot copy file - Source and Destination files are the same"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf

    'validate and/or build the required directory structure for the dest file
    If Not Validate_Build_DestPath%(spDFN$) Then
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub                                        'build failed - exit now
    EndIf
    
    CHDIR saveCWD$                                    'restore current path
    
    'open destination file
    ON ERROR SKIP 1
    OPEN spDFN$ FOR OUTPUT AS #2
    If MM.ERRNO <> 0 Then                             'file open failed - report error
      ? "Open destination file '";fullDFN$;"' failed - ";
      ? MM.ERRMSG$
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf
    
    'copy all data from source file to destination file -
    'majority of the data will be copies in 255 character blocks,
    'the last block may be less that 255 characters and will be
    'copied in the actual size of the remaining data
    f_len%=LOF(#1)                                    'get length of file in bytes
    e%=Choice(f_len%>255,255,f_len%)                  'setup bytes to be processed on first read
    n%=0
    
    Do
      d_buff$=Input$(e%, #1)                          'read e% bytes from the source file
      Inc n%,e%                                       'increment # of bytes read read
      Print #2,d_buff$;                               'copy to dest file - do NOT add CR/LF to end!!
      If e% < 255 Then Exit Do                        'last block just copied - exit loop now
      e%=Choice((n%+255)>f_len%,f_len%-n%,255)        'calc how many bytes to read for next block
      If e% = 0 Then Exit Do                          'reached end of file - exit loop now (file len multiple of 255)
    Loop
    
    'all done - close files and report results
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    
    ? n%;" bytes copied from ";fullSFN$;" to ";fullDFN$
    
  End Sub   'FileBinCopy()

  
  '==========================================================================================================
  '
  '       BINARY FILE LIST Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================

  Sub FileBinList(spSFN$,ipStrt%)
  
    Local a%,e%,f_len%,ln_max%,n%,o%,pm%,pn%,cp$,d_buff$,s$,saveCWD$
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spSFN$,fullSFN$,cp$)                     'format file name & build full file name
    If Not OpenSFN%(spSFN$,fullSFN$) Then             'source file failed to open -
      Exit Sub                                        'exit now
    EndIf
    If ipStrt% >= LOF(#1) Then                        'starting point past end of file
      ? "Error - starting byte offset is past End of File (max valid offset: ";Str$(LOF(#1)-1);")" 
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub                                        'exit now
    EndIf
    
    'since there is currently no way to correctly determine the OPTION DISPLAY settings
    'applying to the console, I am just using 18 lines * 40 chars per line for formatting
    'console output.
    '18*40 is the correct setting for a ILI9314 LCD using OPTION LCDPANEL CONSOLE 1. 
    'Note that ln_max% is 18 - 2, reserving 1 line for the "Press any key to continue..." msg 
    'plus 1 extra to cater for the way the console line scrolling seems to work.
    
    'ch_max%=40                                       'ch_max% is not used in this routine
    ln_max%=16                                        'working max lines value
    f_len%=LOF(#1)                                    'get length of file in bytes
    o%=ipstrt%                                        'set starting position in file
    e%=Choice((o%+8)>f_len%,f_len%-o%,8)              'calc how many bytes to read for first line
    Seek #1,(o%+1)                                    'position file to requested byte offset
    'NOTE: Start position is a byte offset which is 0 based, but bytes in the 
    '===== file are numbered from 1, hence adding 1 to the Seek position above.
    
    Do
      ?                                               'display blank line before listing starts
      For pn% = 1 To ln_max% Step 1
        s$=""                                         'clear string used for alpha display part
        d_buff$=Input$(e%, #1)                        'read e% (up to 8) bytes from the source file
        ? Hex$(o%,5);" ";                             'display hex byte offset for this line     
        For pm% = 1 To e%
          a%=Asc(Mid$(d_buff$,pm%,1))                 'get ASCII value for a single char from d_buff$
          ? Hex$(a%,2);" ";                           'display Hex value
          If a% > 31 and a% < 127 Then
            s$=s$ + Chr$(a%)                          'add printable char to s$
          Else
            s$=s$ + "."                               'else add period (&H2E) to s$ for non-printable char
          EndIf
        Next pm%
        Inc n%,e%                                     'count bytes actually displayed
        Inc o%,e%                                     'increment byte offset by e%
        Do While e% < 8                               'need to pad hex display part for short last line
          ? "   ";
          Inc e%
        Loop
        ? " ";s$                                      'add the alpha part and display the line
        e%=Choice((o%+8)>f_len%,f_len%-o%,8)          'calc how many bytes to read for next line
        If e% < 1 Then Exit Do                        'reached end of file - exit loop now
      Next pn%
      If WaitContinue%(1) Then                        '- need to wait for ok to continue or Esc to quit
        Exit Do
      EndIf
    Loop

    ?                                                 'display blank line
    ? n%;" bytes listed from '";fullSFN$;"'."
    'all done - close file
    CLOSE #1
    
  End Sub   'FileBinList()
  
  
  '==========================================================================================================
  '
  '       BINARY FILE TRANSFER Procedure - designed to be used standalone from the command line.
  '
  '==========================================================================================================
  
  Sub FileBinXfer(spSFN$,spDFN$)
    'This routine is the same as FileBinCopy, except that the destination file is on a different SD Card.
    'This is achieved by first copying the source file to memory, then prompting the user to swap the
    'SD Card for the destination one and entering 'OK' when done, then the file is copied from memory onto 
    'the second SD Card.
    'The filesize that can be accomodated by this function is 102,000 bytes (255 * 400).
    'Both Source File Name and Destination File Name can include Path specifications.
    'The Destination File Name MUST include the full path - i.e. starting from the root directory (A:/)
    'If the Path specified for the Destination File does not exist, it will be automatically created.
        
     Local e%,f_len%,n%,m%,cp$,d_buff$,fullSFN$,fullDFN$
     Local d_array$(400)                              '400 element string array (255 bytes per element)
    
    'ensure files to be used are closed
    ON ERROR SKIP 1
    CLOSE #1
    ON ERROR SKIP 1
    CLOSE #2
    
    cp$=GetCurrPath$(saveCWD$)
    FormatFN(spSFN$,fullSFN$,cp$)                     'format file name & build full file name
    If Not OpenSFN%(spSFN$,fullSFN$) Then             'source file failed to open -
      Exit Sub                                        'exit now
    EndIf
    
    If LOF(#1) < 1 Then
      ? "Cannot copy file - Source file is empty"
      ON ERROR SKIP 1
      Close #1
      Exit Sub
    EndIf
    
    If LOF(#1) > 102000 Then
      ? "Cannot copy file - Source file is too big (max 102000 bytes)"
      ON ERROR SKIP 1
      Close #1
      Exit Sub
    EndIf
        
    FormatFN(spDFN$,fullDFN$,cp$)
    
    If Len(spDFN$) < 1 Then
      ? "Cannot copy file - Destination File Name is blank"
      ON ERROR SKIP 1
      CLOSE #1
      Exit Sub
    EndIf

    If UCase$(Left$(spDFN$,3)) <> "A:/" Then          'dest file requires full path
      '           1         2         3         4 
      '  1234567890123456789012345678901234567890
      ? "Cannot copy file - Destination file"
      ? "name must specify the full path"
      ? "starting with 'A:/'."
      ON ERROR SKIP 1
      Close #1
      Exit Sub
    EndIf
    
    'copy all data from source file to destination file -
    'majority of the data will be copies in 255 character blocks,
    'the last block may be less that 255 characters and will be
    'copied in the actual size of the remaining data
    f_len%=LOF(#1)                                    'get length of file in bytes
    e%=Choice(f_len%>255,255,f_len%)                  'setup bytes to be processed on first read
    n%=0
    o%=0
    
    Do
      d_buff$=Input$(e%, #1)                          'read e% bytes from the source file
      Inc o%,e%                                       'increment # of bytes read so far
      Inc n%                                          'increment read block #
      d_array$(n%)=d_buff$                            'store data read in memory
      If e% < 255 Then Exit Do                        'last block copied - exit loop now
      e%=Choice((o%+255)>f_len%,f_len%-o%,255)        'calc how many bytes to read for next block
      If e% = 0 Then Exit Do                          'reached end of file - exit loop now (file len multiple of 255)
    Loop
    
    ? Str$(o%);" bytes read and stored in RAM."
    
    'reading all done - close file #1
    ON ERROR SKIP 1
    CLOSE #1
    
    'now request the user to swap SD Cards
    '           1         2         3         4 
    '  1234567890123456789012345678901234567890
    ? "Please remove the current SD Card and"
    ? "insert the SD Card the file is to be"
    ? "copied onto."
    WaitOK()                                          'wait for this to be completed
   
    'validate and/or build the required directory structure for the dest file
    If Not Validate_Build_DestPath%(spDFN$) Then
      Exit Sub                                        'build failed - exit now
    EndIf
    
    'CWD was reset to root directory when the SD Card was swapped,
    'however it may have been altered during the Validate_Build_DestPath%() 
    'function (above), so reset it back to the root directory again.
    CHDIR "A:/"
    
    'open destination file - is specified with a full path.
    ON ERROR SKIP 1
    OPEN spDFN$ FOR OUTPUT AS #2
    If MM.ERRNO <> 0 Then                             'file open failed - report error
      ? "Open destination file '" ;spDFN$;"' failed - ";
      ? MM.ERRMSG$
      Exit Sub
    EndIf
    
    For m%=1 to n%
      Print #2,d_array$(m%);                          'copy to dest file - do NOT add CR/LF to end!!
    Next m%
    
    o%=LOF(#2)                                        'get filesize of dest file
    'all array elements copied to destination file
    ON ERROR SKIP 1
    CLOSE #2
    
    ? Str$(o%);" bytes copied from ";fullSFN$;" to ";spDFN$;
    ? " (different SD Card)"

  End Sub   'FileBinXfer()
  

  '==========================================================================================================
  '
  '       WAITCONTINUE Procedure - used by FileList, FileBinList routines and if FileUtils program is RUN.
  '
  '==========================================================================================================
  
  Function WaitContinue%(ipQuit%)
    'screen full, display "Press any key to continue..." message 
    'and wait until a key has been pressed before returning.
    'If ipQuit% is True and the 'Esc' key is pressed, the return
    'value wiil be True (1)to indicate the user requested to Quit.
    'Any other key will return False (0).
    
    Local in_char$
    
    WaitContinue%=0                                   'default return value
    in_char$ = Input$(255, #0)                        'empty console input buffer
    If ipQuit% Then
      ? "Press any key to continue, ESC to Quit."     'display message
    Else
      ? "Press any key to continue..."                'display message
    EndIf
    Do                                                'loop until a key is pressed
      If Loc(#0) > 0 Then
        If ipQuit% And Asc(Input$(1, #0)) = 27 Then   'ESC key    
          WaitContinue%=1                             'return True (1)
        EndIf
        Exit Do                                       'exit loop, a key has been pressed
      EndIf
    Loop 
    in_char$ = Input$(255, #0)                        'empty console input buffer again

  End Function    'WaitContinue%()


  '==========================================================================================================
  '
  '       WAITOK Procedure - used by FileBinXfer routine.
  '
  '==========================================================================================================
  
  Sub WaitOK()
    'waiting for SD Card swap, display "Enter OK when SD Card swap is complete." message 
    'then wait for OK to be entered. OK input is NOT case sensitive
    
    Local in_char$,O_Done%,K_Done% 
    
    in_char$=Input$(255, #0)                          'empty console input buffer
    in_char$=""
    ? "Enter OK when SD Card swap is complete."       'display 'Enter OK when SD Card swap is complete.' msg
    Do 
      If O_Done% = 0 Then
        'get 1 char from console input, check for 'O', set flag when found  
        If UCase$(Input$(1,#0)) = "O" Then O_Done%=1
        
      Else                                            'now looking for 'K'
        'get 1 char from console input, check for 'K', set flag when found  
        If UCase$(Input$(1,#0)) = "K" Then K_Done%=1
      EndIf
    Loop Until O_Done% = 1 AND K_Done% = 1            'loop until O and K have been entered
    in_char$=Input$(255, #0)                          'empty console input buffer again

  End Sub   'WaitOK()

  
  '==========================================================================================================
  '
  '       REPLACE Function - used by various routines (can also be used directly from the Command Prompt)
  '
  '==========================================================================================================
  
  Function Replace$(spString$,spTarg$,spRepl$)
    'scans the spString replacing each instance of the spTarg sequence with the spRepl string.
    'Replacement is case sensitive and spTarg and spRepl do not need to be the same length.
    'ALL instances of spTarg are replaced.
    'NOTE: The is no protection for the replaced string length exceeding 255 chars when replacing
    '----- a small string with a larger one. 
    
    Local p%,t%,tl%,rl%,work$
    
    Replace$=spString$                                'default result - return input string unchanged
    If Len(spTarg$) = 0 Then  
      'not valid - just return now
      Exit Function
    EndIf
    If spTarg$ = spRepl$ Then
      'target and replacement are the same - no change, just return now
      Exit Function
    EndIf
    If Len(spString$) < Len(spTarg$) Then 
      'if spTarg is longer than spString there can never
      'be a match, so just return spString unchanged.
      Exit Function
    EndIf
    work$=spString$                                   'don't modify the spString argument value
    p%=1                                              'start scan at left of string
    tl%=Len(spTarg$)                                  'save length of target text 
    rl%=Len(spRepl$)                                  'save length of replacement text
    Do
      t%=Instr(p%,work$,spTarg$)                      'look for match on spTarg
      If t% = 0 Then Exit Do                          'no match - exit loop
      work$=Left$(work$,(t%-1))+ spRepl$ + Mid$(work$,(t%+tl%)) 'match - replace this instance of spTarg
      p%=t%+rl%                                       'starting position for next target scan
    Loop 
    Replace$=work$                                    'return modified input string 
    '                                                 '(only modified if any matches found)
  End Function   'Replace$()

  
  '==========================================================================================================
  '
  '       TRIM Function - used by various routines (can also be used directly from the Command Prompt)
  '
  '==========================================================================================================

  Function Trim$(spString$)
    'removes any leading and/or trailing spaces from the specified spString.
    
    Local i%,work$=spString$

    For i%=1 To Len(work$)
      If Mid$(work$,i%,1) <> " " Then                 'first non-space character
        Exit For
      EndIf
    Next i%
        
    If i% > Len(work$) Then                           'spString is all spaces
      work$=""
    ElseIf i% > 1 Then                                'leading space(s) exist
      work$=Mid$(work$,i%)                            'remove them
    EndIf
    
    If Len(work$) > 0 Then                           
      For i%=Len(work$) To 1 Step -1
        If Mid$(work$,i%,1) <> " " Then               'last non-space character
          If i% < Len(work$) Then                     'trailing space(s) exist
            work$=Left$(work$,i%)                     'remove them
          EndIf
          Exit For
        EndIf
      Next i%
    EndIf
    
    Trim$=work$                                       'return string
    
  End Function    'Trim$()
  
  
  '==========================================================================================================
  '
  '       GETCURRPATH$ Function - used by most routines
  '
  '==========================================================================================================
  
  Function GetCurrPath$(spCWD$)
    'gets the current path specification (via CWD$ command) and returns it in a format required for use by
    'other functions - i.e. with a trailing forward slash.
    'also sets the spCWD$ argument variable to the CWD$ value modified to include the Logical Drive Letter
    'so that it will be in a format that can be used directly to restore the CWD$ setting if reqd.
    Local cp$
                
    spCWD$=Replace$(CWD$,"\","/")
    If Left$(spCWD$,1) = ":" Then                     'root directory needs Logical Drive Letter
      spCWD$="A" + spCWD$                             'add it  
    EndIf
    
    cp$=spCWD$                                        'build Current Path return value
    If Right$(cp$,1) <> "/" Then
      cp$=cp$ + "/"                                   'ensure current path includes a trailing forward slash
    EndIf
    
    GetCurrPath$=cp$                                  'return current path

  End Function    'GetCurrPath$()


  '==========================================================================================================
  '
  '       FORMAT FILE NAME Routine - used by many of the routines
  '
  '==========================================================================================================
 
  Sub FormatFN(spFN$,spfullFN$,spcp$)
    'ensure the format of the specified File Name (including any path specification) is as expected. 
    'any required modifications are done directly on and returned via the spFN$ argument.
    'also builds the full File Name including the absolute path. This is returned via the spfullFN$
    'argument variable. The spcp$ argument is used to pass in the current path which is needed when
    'building the full File Name.
    
    spfullFN$=""
    spFN$=Trim$(spFN$)
    If Len(spFN$) = 0 Then Exit Sub                   'specified file name is blank
 
    spFN$=Replace$(spFN$,"\","/")                     'replace back slashes with forward slashes
    If Left$(spFN$,1) = "/" Then                      
      spFN$=Mid$(spFN$,2)                             'remove a leading slash (if present)
    EndIf
    If Left$(spFN$,1) = ":" Then                      'root directory needs Logical Drive Letter
      spFN$="A" + spFN$                               'add it
    EndIf
    'build full File Name
    spfullFN$=Choice(Instr(spFN$,":")=0,spcp$+spFN$,spFN$)
    
  End Sub 

  
  '==========================================================================================================
  '
  '       OPENSFN% Function - used by FileExtract, FileList, FileBinCopy, FileBinList & FileBinXfer routines
  '
  '==========================================================================================================
  
  Function OpenSFN%(spSFN$,spfullSFN$)
    'Attempts to open the specified source file for INPUT as fnbr# 1.
    'Any validation errors are reported via a console message 
    'and the return value will be False (0).
    'If the file is opened successfully the return value will be True (1).

    OpenSFN%=0                                        'default return value - false, open failed
    
    If spSFN$ = "" Then
      ? "Cannot open the source file - File Name is blank."
      Exit Function
    EndIf

    'attempt to open the specified file.
    ON ERROR SKIP 1
    OPEN spSFN$ FOR INPUT AS #1
    If MM.ERRNO <> 0 Then
      ? "Source file '";spfullSFN$;"' cannot be found"
      Exit Function
    EndIf
    ON ERROR CLEAR
    
    OpenSFN%=1                                        'set return value = True - open successful

  End Function    'OpenSFN%()
  
  
  '==========================================================================================================
  '
  '       VALIDATE or BUILD DESTINATION PATH Function - used by many of the routines
  '
  '==========================================================================================================
  
  Function Validate_Build_DestPath%(spDFN$)
    'Validate and/or build the directory structure required for the destination file.
    'If an error is encountered, it will be reported via a console message 
    'and the return value will be False (0).
    'If the required directory structure validated or built successfully 
    'the return value will be True (1).
    
    Local i%,k%,n%,md%,c$,dp$
    
    Validate_Build_DestPath%=0                        'default return value - error encountered
    
    'check destination path (if specified) and
    'create any directories that do not currently exist 
    n%=Len(spDFN$)
    k%=1    
    
    FOR i%=1 TO n%
      c$=Mid$(spDFN$,i%,1) 
      'check for a fwd or back slash
      If c$ = "/" Then
        dp$=Mid$(spDFN$,k%,(i%-k%))                   'get next directory name in path spec
        k%=i%+1                                       'save pointer to char after slash
        If UCase$(dp$) = "A:" Then
          'root directory needs trailing slash for CHDIR, other directory levels don't
          dp$=dp$ + "/"                               'add trailing slash
        EndIf
        Do
          'try to change to this directory
          ON ERROR SKIP 1
          CHDIR dp$
          If MM.ERRNO = 0 Then                        'success - directory must exist
            ON ERROR CLEAR 
            md%=0                                     'reset MKDIR done flag             
            EXIT DO                                   'continue scan for more directory levels
          EndIf
          'CHDIR failed
          If md% Then                                 'this directory was just created ???
            ? "Change Directory to '" + dp$;          '- report fatal error
            ? "' after creating it failed - "; 
            ? MM.ERRMSG$
            Exit Function                             'return False - error encountered
          EndIf
          '- need to create this directory
          md%=0                                       'reset MKDIR done flag
          ON ERROR SKIP 1
          MKDIR dp$                                   'create directory
          If MM.ERRNO <> 0 Then                       'MKDIR failed - report fatal error
            ? "Create directory '" + dp$ + "' failed - ";
            ? MM.ERRMSG$
            Exit Function                             'return False - error encountered
          EndIf
          md%= 1                                      'set MKDIR done flag
        Loop
      EndIf
    Next i%
    
    'if we get here the directory structure either already existed, or was built successfully
    Validate_Build_DestPath%=1                        'return True - success
    
  End Function 
  
  
  '==========================================================================================================
  '
  '       READ NEXT RECORD Function - used to manage record based file input
  '
  '==========================================================================================================
  
  Function ReadNextRecord%(spfullFN$,spRec$)
    'attempts to read the next record from fnbr# 1 (source file).
    'if an error is encountered, error details are reported via the console.
    'all files are attempted to be closed so that the program can exit cleanly
    'and the function returns False (0).
    'if the next record is read successfully, the function returns True (1) 
    'and the record data is passed back via the spRec$ argument variable.

    ON ERROR SKIP 1
    LINE INPUT #1,spRec$                            'read next record
    If MM.ERRNO <> 0 Then                           'read failed    
      ? "Read source file failed - ";MM.ERRMSG$
      ? "File '";spfullFN$;"' may be a stream file."
      ON ERROR SKIP 1
      CLOSE #1
      ON ERROR SKIP 1
      CLOSE #2
      ON ERROR SKIP 1
      CLOSE #3
      ReadNextRecord% = 0                           'return False - failed result
    Else
      ReadNextRecord% = 1                           'return True - successful
    EndIf
    
  End Function    'ReadNextRecord%()
